import logging
from typing import List, Literal, Optional, Tuple, Union
from ase import Atoms
from ase.geometry.rdf import get_rdf
import numpy as np
import numpy.typing as npt

import psutil
import torch

logger = logging.getLogger("franken")


def batched_distance_matrix(traj: list[Atoms], num_cells=1) -> np.ndarray:
    """From 'forces are not enough', computes distance matrix between atom positions
    dealing with non-cubic cell.

    TODO: This is likely to fail in some cases as dealing with periodic systems
          is less robust than ASE! We could switch to doing this with ase `get_distances`

    TODO: This only works if using PBC on all 3 axes
    """

    ### Batching calculations
    def max_ram(traj_len):
        # Final multiplier should be *3, becomes *4 to include reduced dist and *5
        # for safety buffer.
        return 4 * traj_len * len(traj[0]) * len(traj[0]) * (num_cells * 2 + 1) ** 3 * 5

    if torch.cuda.is_available():
        avail_ram = torch.cuda.mem_get_info()[0] * 0.9
        dev = "cuda"
    else:
        avail_ram = psutil.virtual_memory().available * 0.4
        dev = "cpu"
    batch_size = len(traj)
    while max_ram(batch_size) > avail_ram:
        logger.debug(
            f"Available {avail_ram / 2 ** 30:.2f}GiB not enough "
            f"for batch size {batch_size} (needs "
            f"{max_ram(batch_size) / 2 ** 30:.2f}GiB)"
        )
        batch_size //= 2
    logger.debug(f"Chosen batch size {batch_size}")

    ### Batch-invariant quantities
    # Extract all atom positions: [N, A, 3]
    coords = (
        torch.stack([torch.from_numpy(a.get_positions(wrap=True)) for a in traj], dim=0)
        .float()
        .to(device=dev)
    )
    # Cell is [3, 3] containing size of unit cell. cells is [N, 3, 3]
    cells = torch.stack(
        [torch.from_numpy(frame.cell.array).float() for frame in traj], 0
    ).to(device=dev)
    # If num_cells = 1, pos = -1, 0, 1: the shifts of the unit cell to be
    # considered. Call the number of positions P (in the example above P=3)
    pos = torch.arange(-num_cells, num_cells + 1, 1, device=dev)
    # combos is all possible combinations of `pos` in the 3 xyz dimensions. [P^3, 3].
    # for example there will be [-1, -1, -1], [-1, 0, -1], [1, 1, 1], etc
    combos = (
        torch.stack(torch.meshgrid(pos, pos, pos, indexing="xy"))
        .permute(3, 2, 1, 0)
        .reshape(-1, 3)
        .to(cells.device)
    )
    # and shifts is the actual shifts in cell units
    # cells * combos is [N, P^3, 3, 3],  shifts is [N, P^3, 3]
    shifts = torch.sum(cells[:, None, :, :] * combos[None, :, :, None], dim=-1)

    ### Compute batched distances
    all_dist = []
    for batch_start in range(0, len(traj), batch_size):
        # b_coords: [N, A, 3]
        b_coords = coords[batch_start : batch_start + batch_size]
        b_shifts = shifts[batch_start : batch_start + batch_size]
        # the coordinates get shifted by each of the shift: [N, A, P^3, 3]
        shifted = b_coords[:, :, None, :] + b_shifts[:, None, :, :]
        # Euclidean distance between atom positions and all shifted atom positions
        # [N, A, A, P^3, 3] -> [N, A, A, P^3]
        dist = b_coords.unsqueeze(2).unsqueeze(2) - shifted.unsqueeze(1)
        dist = torch.linalg.vector_norm(dist, ord=2, dim=-1)
        # Minimum distance among all shifted positions (Minimum-image convention)
        # [N, A, A]
        dist = dist.min(dim=-1)[0]
        all_dist.append(dist)
    return torch.cat(all_dist, 0).numpy(force=True)


def ase_distance_matrix(traj: list[Atoms]) -> npt.NDArray[np.float32]:
    dist: List[npt.NDArray[np.float32]] = []
    for config in traj:
        dist.append(config.get_all_distances(mic=True).astype(np.float32))
    return np.stack(dist, 0)


def distance_matrix(
    traj: list[Atoms], version: Literal["ase", "fast"]
) -> npt.NDArray[np.float32]:
    if version == "ase":
        return ase_distance_matrix(traj)
    elif version == "fast":
        return batched_distance_matrix(traj)
    else:
        raise RuntimeError(
            f"distance matrix implementation '{version}' not understood."
        )


def fast_avg_rdf(
    traj: list[Atoms],
    rmax: float,
    nbins: int,
    no_dists: bool = False,
    distance_matrix: np.ndarray | None = None,
    elements=None,
) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    if distance_matrix is None:
        distance_matrix = batched_distance_matrix(traj)
    cell_volumes = np.array([frame.cell.volume for frame in traj])
    bins = np.linspace(0, rmax, nbins + 1)

    # atom-types must be uniform throughout the trajectory, otherwise use 'ase' version
    atom_types = traj[0].numbers
    for frame in traj:
        if not (atom_types == frame.numbers).all():
            raise NotImplementedError(
                "Trajectory does not have uniform atomic types or number of atoms. "
                "This cannot be handled by the fast RDF calculator, please use the 'ase' RDF version."
            )

    if elements is not None:
        # Fetch the portion of the distance matrix which concerns the two
        # chosen atomic types
        indices0 = atom_types == elements[0]
        indices1 = atom_types == elements[1]
        distance_matrix = distance_matrix.transpose(1, 2, 0)[
            np.ix_(indices0, indices1)
        ].transpose(2, 0, 1)

    rep_cell_volumes = np.repeat(
        cell_volumes, distance_matrix.shape[1] * distance_matrix.shape[2]
    )
    flat_pdist = distance_matrix.flatten()
    data_shape = flat_pdist.shape[0]

    nz_dist_mask = flat_pdist != 0
    rep_cell_volumes = rep_cell_volumes[nz_dist_mask]
    flat_pdist = flat_pdist[nz_dist_mask]
    data_hist, _ = np.histogram(flat_pdist, bins, weights=rep_cell_volumes)

    rho_data = data_shape
    Z_data = rho_data * 4 / 3 * np.pi * (bins[1:] ** 3 - bins[:-1] ** 3)
    rdf = data_hist / Z_data
    if no_dists:
        return rdf
    return rdf, np.arange(float(rmax / nbins) / 2, rmax, float(rmax / nbins))


def ase_avg_rdf(
    traj: list[Atoms],
    no_dists=False,
    distance_matrix: np.ndarray | None = None,
    elements=None,
    **kwargs,
) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    """Calculate average RDF with ASE."""
    dists = None
    rdf_list = []
    for i, atoms in enumerate(traj):
        # NOTE: `get_rdf` is a very slow way to compute a histogram
        #       as it loops over each bin in python. We might consider
        #       changing this if performance is a problem.
        rdf, dists = get_rdf(
            atoms,
            distance_matrix=distance_matrix[i] if distance_matrix is not None else None,
            no_dists=False,
            elements=elements,
            **kwargs,
        )
        rdf_list.append(rdf)
    avg_rdf = np.mean(rdf_list, axis=0)
    if no_dists:
        return avg_rdf
    return avg_rdf, dists  # type: ignore


def avg_rdf(
    traj: list[Atoms],
    rmax: float,
    nbins: int,
    no_dists: bool = False,
    distance_matrix: np.ndarray | None = None,
    elements: Optional[Tuple[int, int]] = None,
    version: Literal["ase", "fast"] = "fast",
) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    if version == "ase":
        return ase_avg_rdf(
            traj,
            rmax=rmax,
            nbins=nbins,
            no_dists=no_dists,
            distance_matrix=distance_matrix,
            elements=elements,
        )
    elif version == "fast":
        return fast_avg_rdf(
            traj,
            rmax=rmax,
            nbins=nbins,
            no_dists=no_dists,
            distance_matrix=distance_matrix,
            elements=elements,
        )
    else:
        raise RuntimeError(f"average RDF implementation '{version}' not understood.")


def compute_rdf_mae(rdf: np.ndarray, ref_rdf: np.ndarray, delta_r: float):
    """MAE between RDFs"""
    return np.sum(np.abs(rdf - ref_rdf)) * delta_r


def stability_rdf(
    traj: list[Atoms],
    dist: np.ndarray,
    ref_rdf: np.ndarray,
    rmax: float,
    nbins: int,
    threshold=1.0,
    stability_interval=20,
    elements: Optional[Tuple[int, int]] = None,
    rdf_version: Literal["fast", "ase"] = "fast",
):
    """Calculate RDF stability metrics between a test trajectory and a reference trajectory

    The test RDF is computed on a sliding window along the test trajectory, and compared
    using the MAE metric to the full reference RDF. If the threshold is exceeded, the test
    trajectory is deemed unstable and computations are interrupted.
    """
    stability_mae = []
    unstable_time = len(traj)  # index of first point which is unstable
    # NOTE: This was changed from computing stability once every `stability_interval`
    #       to computing it every time-step on windows of size `stability_interval`.
    #       this is consistent with 'forces are not enough'.
    for i in range(1, len(traj) - stability_interval):
        current_interval = slice(i, i + stability_interval)
        current_rdf = avg_rdf(
            traj[current_interval],
            distance_matrix=dist[current_interval],
            rmax=rmax,
            nbins=nbins,
            no_dists=True,
            version=rdf_version,
            elements=elements,
        )
        assert isinstance(current_rdf, np.ndarray)
        current_mae = compute_rdf_mae(current_rdf, ref_rdf, rmax / nbins)
        stability_mae.append(current_mae)
        if current_mae > threshold:
            unstable_time = i
            logger.warning(f"Trajectory becomes unstable at time {i}")
            break
    return unstable_time, stability_mae



import inspect
import logging
import os
import time
from pathlib import Path
from typing import Any, Union

import hydra
import numpy as np
import torch
from omegaconf import DictConfig

import ase
import ase.calculators
import ase.calculators.calculator
import ase.io
import ase.io.trajectory
import ase.md.md
import ase.units
from franken.calculators.ase_calc import FrankenCalculator
from franken.backbones.utils import CacheDir, get_checkpoint_path, load_model_registry
from franken.mdgen.mdgen_utils import (
    FrankenMDLogger,
    MDError,
    get_md_nan_checker,
    register_hydra_resolvers,
)
from franken.utils.misc import pprint_config, remove_root_logger_handlers, setup_logger


try:
    from mace.calculators import MACECalculator  # type: ignore
except ImportError:
    MACECalculator = None
try:
    from fairchem.core.common.relaxation.ase_utils import OCPCalculator  # type: ignore
except ImportError:
    OCPCalculator = None

logger = logging.getLogger("franken")


def init_calculator(
    calc_type: str,
    model_path: Union[str, Path, None],
    device: Union[int, str, torch.device],
    **calc_kwargs,
) -> ase.calculators.calculator.Calculator:
    """Initialize the calculator based on the specified type from the given path.

    Args:
        calc_type
            Describes the calculator which should be used. Can be either 'franken',
            which loads a Franken model from the specified :code:`model_path`, one
            of the supported backbone families (e.g. "MACE", "fairchem") which use
            the calculators from the respective repositories, or it can also be a
            backbone ID (e.g. "MACE-L0") such that, if :code:`model_path` is not
            specified, the backbone itself is loaded as a model for MD simulations.
        model_path
            Optional path to the backbone used to compute energies and forces for
            the MD simulation.
        device
            Device on which the backbone should run.
        calc_kwargs
            Additional arguments to be passed to the Calculator class which corresponds
            to the chosen :code:`calc_type`. For the supported keyword arguments for
            the "franken" calculator, see :class:`franken.ase.FrankenCalculator`.
    Returns:
        Calculator
    """
    calc: ase.calculators.calculator.Calculator
    if calc_type == "franken":
        if model_path is None:
            raise ValueError(
                f"Model path must not be 'None' for calculator of type '{calc_type}'. "
                "Check your config file and re-run asemd."
            )
        calc = FrankenCalculator(model_path, device=device, **calc_kwargs)
    else:
        registry = load_model_registry()
        backbone_families = set([entry["kind"] for entry in registry.values()])
        if calc_type in backbone_families:
            if model_path is None:
                raise ValueError(
                    f"Model path must not be 'None' for calculator of type '{calc_type}'. "
                    "Check your config file and re-run asemd."
                )
            calc_family = calc_type
        elif calc_type in registry:
            calc_family = registry[calc_type]["kind"]
            if model_path is None:
                logger.info(
                    f"Model path was not specified for calculator of type '{calc_type}'. "
                    f"Using the default model downloaded from '{registry[calc_type]['remote']}'"
                )
                model_path = get_checkpoint_path(calc_type)
        else:
            raise ValueError(
                f"Calculator of type '{calc_type}' is not valid. "
                f"Expected 'franken', a backbone family "
                f"(i.e. one of {backbone_families}) or a valid model ID."
            )
        # Initialize the Fairchem/Mace calculators
        if calc_family == "MACE":
            if MACECalculator is None:
                raise ImportError(
                    f"Failed to initialize calculator of type '{calc_type}' "
                    "because mace could not be imported. Check that mace is "
                    "installed in the current environment."
                )
            calc = MACECalculator(str(model_path), device=str(device), **calc_kwargs)
            # MACE Adds a root-level logger which makes all logs print out duplicates
            remove_root_logger_handlers()
        elif calc_family == "fairchem":
            if OCPCalculator is None:
                raise ImportError(
                    f"Failed to initialize calculator of type '{calc_type}' "
                    "because fairchem could not be imported. Check that fairchem is "
                    "installed in the current environment."
                )
            calc = OCPCalculator(
                checkpoint_path=str(model_path),
                cpu=torch.device(device).type != "cuda",
                **calc_kwargs,
            )
            calc.trainer.model.float()  # type: ignore
        else:
            raise NotImplementedError(
                f"Calculator for backbone family '{calc_family}' is not implemented."
                " Only 'MACE' and 'fairchem' families have an implemented calculator."
            )
    return calc


def cfg_fn_has_param(cfg, param_name):
    import hydra._internal.instantiate._instantiate2

    init_v_func = hydra._internal.instantiate._instantiate2._resolve_target(
        cfg._target_, ""
    )
    return param_name in inspect.signature(init_v_func).parameters.keys()


def run_md_simulation(
    output_dir: Path,
    integrator_cfg,
    initial_velocity_cfg,
    initial_configuration: ase.Atoms,
    num_steps: int,
    save_every: int,
    log_every: int,
    seed: int,
    output_format="traj",
):
    """Run MD simulation for given atoms at a specific temperature."""
    if num_steps == 0:
        return output_dir / f"replica_{seed}.{output_format}"

    # Set random seed
    torch.manual_seed(seed)
    np.random.seed(seed)

    # Initialize velocities/conditions if given.
    ## assert (initial_configuration.get_momenta() == 0).all()
    # Extra parameters to velocity initialization always include the
    # initial configuration. Sometimes also a specific rng (not all
    # velocity inits are random!)
    extra_params: dict[str, Any] = {"atoms": initial_configuration}
    if initial_velocity_cfg is not None:
        if cfg_fn_has_param(initial_velocity_cfg, "rng"):
            extra_params["rng"] = np.random.default_rng(seed=seed)
            hydra.utils.call(initial_velocity_cfg, **extra_params)

    # Initialize trajectory
    trajectory_path = output_dir / f"replica_{seed}.traj"
    if trajectory_path.exists():
        logger.warning(
            f"Trajectory path '{trajectory_path.resolve()}' already exists and will be overwritten."
        )
    trajectory = ase.io.Trajectory(trajectory_path, "w", initial_configuration)
    assert isinstance(trajectory, ase.io.trajectory.TrajectoryWriter)

    # MD Setup and running
    logger.info(f"{num_steps} steps of MD simulation starting...")
    t_start = time.time()
    integrator = hydra.utils.instantiate(
        integrator_cfg,
        atoms=initial_configuration,
    )
    md_logger = FrankenMDLogger(
        dynamics=integrator,
        atoms=initial_configuration,
        stress=False,
        peratom=False,
    )
    # nan checker needs to go first to be effective.
    integrator.attach(get_md_nan_checker(atoms=initial_configuration), interval=1)
    integrator.attach(trajectory.write, interval=save_every)
    integrator.attach(md_logger, interval=log_every)
    try:
        integrator.run(num_steps)
        t_end = time.time()
        logger.info(f"{num_steps} simulation steps complete in {t_end - t_start:.2f}s.")
    except MDError as e:
        t_end = time.time()
        logger.error(f"Simulation failed after {t_end - t_start:.2f}s.", exc_info=e)

    if output_format != "traj":
        logger.info(f"Converting ASE traj to the specified format {output_format}")
        traj = ase.io.read(trajectory_path, index=":")
        new_trajectory_path = output_dir / f"replica_{seed}.{output_format}"
        if new_trajectory_path.exists():
            logger.warning(
                f"Trajectory path '{new_trajectory_path.resolve()}' already exists and will be overwritten."
            )
        ase.io.write(new_trajectory_path, traj)
        os.remove(trajectory_path)
        return new_trajectory_path
    return trajectory_path


@hydra.main(version_base="1.3", config_path="configs", config_name="md_run.yaml")
def md_main(cfg: DictConfig):
    output_dir = Path(cfg.paths.run_dir)
    output_dir.mkdir(parents=True, exist_ok=True)

    if torch.cuda.is_available():
        device = torch.device(torch.cuda.current_device())
    else:
        device = torch.device("cpu")

    logging_level = cfg.console_logging_level.upper()
    setup_logger(level=logging_level, directory=output_dir, logname="asemd")
    pprint_config(cfg)

    CacheDir.initialize()

    # Load initial configuration
    init_config_index = cfg.get("init_config_index", 0)
    init_config = ase.io.read(str(cfg.init_config_path), index=init_config_index)
    assert isinstance(init_config, ase.Atoms)  # check we didn't read list[Atoms]

    # if the system has hydrogen atoms, set their mass according to config (default=1 a.u.)
    if (init_config.get_atomic_numbers() == 1).any() and "hydrogen_mass" in cfg:
        mass_H = cfg.get("hydrogen_mass")
        masses = init_config.get_masses()
        init_config.get_masses()[init_config.get_atomic_numbers() == 1] = mass_H
        init_config.set_masses(masses)

        logger.info(f"Setting hydrogen mass to {int(mass_H)} a.u.")

    # Initialize calculator
    calc = init_calculator(**cfg.calculator, device=device)

    # Setup a few MD parameters
    num_md_steps = int(round(cfg.md_length_ns * 1e6 / cfg.timestep_fs))

    # Get seeds from configs
    seeds = [cfg.seed] if isinstance(cfg.seed, int) else cfg.seed

    # Run MD
    traj_paths = []
    for i, seed in enumerate(seeds):
        rep_init = init_config.copy()
        rep_init.calc = calc
        saved_path = run_md_simulation(
            output_dir=output_dir,
            integrator_cfg=cfg.integrator,
            initial_velocity_cfg=cfg.initial_velocity,
            initial_configuration=rep_init,
            num_steps=num_md_steps,
            output_format=cfg.output_format,
            save_every=cfg.save_every,
            log_every=cfg.log_every,
            seed=seed,
        )
        traj_paths.append(saved_path)
        logger.info(f"Replica {i} ({seed=}, {cfg.temperature_K=:.1f}K) finished")
        logger.info(f"Run output can be found at '{saved_path.resolve()}'")
        print()  # just for nicer output


def md_cli_entry_point():
    register_hydra_resolvers()
    md_main()


if __name__ == "__main__":
    md_cli_entry_point()



import torch

from franken.metrics.base import BaseMetric
from franken.metrics.functions import *  # noqa: F403
from franken.metrics.registry import registry


__all__ = ["registry"]


def available_metrics() -> list[str]:
    return registry.available_metrics


def register(name: str, metric_class: type) -> None:
    registry.register(name, metric_class)


def init_metric(
    name: str, device: torch.device, dtype: torch.dtype = torch.float32
) -> BaseMetric:
    return registry.init_metric(name, device, dtype)



import logging
import weakref
from typing import Any

from omegaconf import OmegaConf
import ase.md.md
import ase.units
from ase import Atoms
import numpy as np

from franken.mdgen.tm23_utils import tm23_temp
from franken import FRANKEN_DIR


logger = logging.getLogger("franken")


class MDError(RuntimeError):
    pass


def get_md_nan_checker(atoms: Atoms):
    def check_nans():
        # Forces are not set as an attribute on the atoms.
        # Therefore we check the positions which should be derived
        # from the forces (and therefore NaNs should be propagated)
        forces = atoms.get_forces()
        if not np.all(np.isfinite(forces)):
            raise MDError("Inf/NaN values encountered in atomic forces")
        energy = atoms.get_potential_energy()
        if not np.all(np.isfinite(energy)):
            raise MDError("Inf/NaN values encountered in atomic energies")

    return check_nans


class FrankenMDLogger:
    def __init__(
        self,
        dynamics: ase.md.md.MolecularDynamics,
        atoms: Atoms,
        stress: bool,
        peratom: bool,
    ):
        self.dynamics = (
            weakref.proxy(dynamics) if hasattr(dynamics, "get_time") else None
        )
        self.atoms = atoms
        global_natoms = atoms.get_global_number_of_atoms()
        self.stress = stress
        self.peratom = peratom
        self.logged_data: list[dict[str, Any]] = []
        self.hdr = ""
        self.fmt = ""
        if self.dynamics is not None:
            self.hdr = f"{'Time[ps]':10s} "
            self.fmt = "{:<10.4f} "
        if self.peratom:
            self.hdr += (
                f"{'Etot/N[eV]':12s} {'Epot/N[eV]':12s} {'Ekin/N[eV]':12s} {'T[K]':6s}"
            )
            self.fmt += "{:<12.4f} {:<12.4f} {:<12.4f} {:<6.1f}"
        else:
            self.hdr += (
                f"{'Etot[eV]':12s} {'Epot[eV]':12s} {'Ekin[eV]':12s} {'T[K]':6s}"
            )
            # Choose a sensible number of decimals
            if global_natoms <= 100:
                digits = 4
            elif global_natoms <= 1000:
                digits = 3
            elif global_natoms <= 10000:
                digits = 2
            else:
                digits = 1
            self.fmt += 3 * f"{{:<12.{digits}f}} " + "{:<6.1f}"

        if self.stress:
            self.hdr += (
                "      ---------------------- stress [GPa] " "-----------------------"
            )
            self.fmt += 6 * " {:<10.3f}"
        logger.info(self.hdr)

    def __call__(self):
        data_item = {}
        epot = self.atoms.get_potential_energy()
        ekin = self.atoms.get_kinetic_energy()
        temp = self.atoms.get_temperature()

        global_natoms = self.atoms.get_global_number_of_atoms()
        if self.peratom:
            epot /= global_natoms
            ekin /= global_natoms
        if self.dynamics is not None:
            t = self.dynamics.get_time() / (1000 * ase.units.fs)
            dat = [
                t,
            ]
            data_item["time"] = t
        else:
            dat = []

        data_item["epot"] = epot
        data_item["ekin"] = ekin
        data_item["temp"] = temp

        dat += [epot + ekin, float(epot), ekin, temp]

        if self.stress:
            dat += list(self.atoms.get_stress(include_ideal_gas=True) / ase.units.GPa)

        self.logged_data.append(data_item)
        logger.info(self.fmt.format(*dat))


def register_hydra_resolvers():
    OmegaConf.register_new_resolver("fs_to_ase", lambda x: float(x) * ase.units.fs)
    OmegaConf.register_new_resolver("invfs_to_ase", lambda x: float(x) / ase.units.fs)
    OmegaConf.register_new_resolver("tm23_temp", tm23_temp)
    # TODO: Not sure this will work when franken is pip-installed
    OmegaConf.register_new_resolver(
        "default_datasets_path",
        lambda: str((FRANKEN_DIR.parent / "datasets").resolve()),
    )



from pathlib import Path
import time
from typing import Literal, Union

import numpy as np
import torch

from ase.calculators.calculator import Calculator, all_changes
import ase.io
from franken.data import BaseAtomsDataset, Configuration
from franken.rf.model import FrankenPotential
from franken.utils.misc import get_device_name


def run(franken_ckpt, device, traj_path, num_reps=100):
    franken = FrankenPotential.load(
        franken_ckpt,
        map_location=device,
        rf_weight_id=None,
    )
    dataset = BaseAtomsDataset.from_path(
        data_path=None,
        split="md",
        gnn_backbone_id=franken.gnn_backbone_id,
    )
    atoms = ase.io.read(traj_path, index=0)

    config_idx = dataset.add_configuration(atoms)  # type: ignore
    cpu_data = dataset.__getitem__(config_idx, no_targets=True)
    assert isinstance(cpu_data, Configuration)
    data = cpu_data.to(device)
    t_iter = []
    for rep in range(num_reps):
        t_s = time.time()
        energy, forces = franken.energy_and_forces(data, forces_mode="torch.autograd")
        energy = energy.cpu()
        if forces is not None:
            forces = forces.cpu()
        t_iter.append(time.time() - t_s)
    print(f"Took {np.mean(t_iter):.2f}s average time")


if __name__ == "__main__":
    franken_ckpt = "/leonardo_work/IIT24_AtomSim/franken_data/experiments/water/multinode_test/test2/run_250108_175820_2f84bf52/best_ckpt.pt"
    device = "cuda:0"
    traj_path = "/leonardo/home/userexternal/gmeanti0/mlpot_transfer/datasets/water/ML_AB_dataset_1.xyz"

    run(franken_ckpt, device, traj_path)




import logging
import os
from typing import Literal, Optional, Tuple, Union

import torch

from franken.backbones.utils import load_checkpoint
from franken.data import Configuration
from franken.rf.atomic_energies import AtomicEnergiesShift
from franken.rf.heads import (
    BiasedOrthogonalRFF,
    Linear,
    MultiScaleOrthogonalRFF,
    OrthogonalRFF,
    RandomFeaturesHead,
    TensorSketch,
)
from franken.rf.scaler import FeatureScaler
from franken.utils.jac import jacfwd, tune_jacfwd_chunksize
from franken.utils.misc import sanitize_init_dict

logger = logging.getLogger("franken")


class FrankenPotential(torch.nn.Module):
    """Maps atomic configurations into high-dimensional vectors through a neural network and random features.

    This class provides both functions to map atom configurations with GNN + random features through methods
    :meth:`~franken.rf.model.FrankenPotential.feature_map` and :meth:`~franken.rf.model.FrankenPotential.grad_feature_map`,
    and to perform inference on the potential function of atom configurations given a learned
    linear model (:meth:`~franken.rf.model.FrankenPotential.energy_and_forces`).

    Args:
        gnn_backbone_id (str):
        kernel_type (str): The kernel type to use. Must be one of :code:`gaussian`, :code:`poly`, or :code:`linear`.
        random_features_params (dict): A dictionary of parameters for the random features head. For each kernel type the default parameters can be found :ref:`below <random_features_params>`.
        interaction_block (int): GNN layer used for feature extraction
        jac_chunk_size (int or "auto"):

    Attributes:
        gnn (torch.nn.Module): The graph neural network which is used for feature extraction, loaded from
            a pretrained checkpoint.
        rf (torch.nn.Module): Random-features module :class:`franken.rf.heads.RandomFeaturesHead`.
    """

    def __init__(
        self,
        gnn_backbone_id: str,
        kernel_type: Literal[
            "gaussian", "poly", "linear", "biased-gaussian", "multiscale-gaussian"
        ],
        random_features_params: dict = {},
        interaction_block=2,
        jac_chunk_size: Union[int, Literal["auto"]] = "auto",
        scale_by_Z: bool = True,
        num_species: int = 1,
        atomic_energies: Optional[dict[int, torch.Tensor | float]] = None,
        **gnn_kwargs,
    ):
        super(FrankenPotential, self).__init__()
        # stored here as simpler to access than through self.gnn
        self.gnn_backbone_id = gnn_backbone_id
        self.kernel_type = kernel_type
        self.jac_chunk_size = jac_chunk_size
        self.num_species = num_species

        # Caches for jacobian function
        self._grad_energy_jacfn = None
        self._grad_fmap_jacfn = None

        # Initialize `gnn`, `rf`, `input_scaler`, `energy_shift` submodules
        self.gnn = load_checkpoint(
            gnn_backbone_id=gnn_backbone_id,
            interaction_block=interaction_block,
            **gnn_kwargs,
        )
        rf_feature_dim = self.gnn.feature_dim()

        rf_cls: type[RandomFeaturesHead]
        if self.kernel_type == "gaussian":
            rf_cls = OrthogonalRFF
        elif self.kernel_type == "poly":
            rf_cls = TensorSketch
        elif self.kernel_type == "linear":
            rf_cls = Linear
        elif self.kernel_type == "biased-gaussian":
            rf_cls = BiasedOrthogonalRFF
        elif self.kernel_type == "multiscale-gaussian":
            rf_cls = MultiScaleOrthogonalRFF
        else:
            raise ValueError(self.kernel_type)

        random_features_params = sanitize_init_dict(rf_cls, random_features_params)
        self.rf = rf_cls(input_dim=rf_feature_dim, **random_features_params)

        self.input_scaler = FeatureScaler(
            input_dim=self.rf.input_dim,
            statistics=None,
            scale_by_Z=scale_by_Z,
            num_species=num_species,
        )
        self.energy_shift = AtomicEnergiesShift(
            num_species=num_species, atomic_energies=atomic_energies
        )

    @property
    @torch.jit.unused
    def hyperparameters(self):
        hps = {
            "franken": {"kernel_type": self.kernel_type, **self.gnn.init_args()},
            "random_features": self.rf.init_args(),
            "input_scaler": self.input_scaler.hyperparameters(),
        }
        return hps

    def save(self, path: os.PathLike | str, multi_weights: torch.Tensor | None = None):
        if multi_weights is not None:
            assert torch.is_tensor(multi_weights)
            assert multi_weights.ndim <= 2
            assert multi_weights.shape[-1] == self.rf.weights.shape[-1]

        ckpt = {
            "kernel_type": self.kernel_type,
            "jac_chunk_size": self.jac_chunk_size,
            "multi_weights": multi_weights,
            "num_species": self.num_species,
            "rf": {
                "config": self.rf.init_args(),
                "state_dict": self.rf.state_dict(),
            },
            "input_scaler": {
                "config": self.input_scaler.init_args(),
                "state_dict": self.input_scaler.state_dict(),
            },
            "energy_shift": self.energy_shift.state_dict(),
            "gnn": {
                "config": {
                    **self.gnn.init_args(),
                },
            },
        }
        torch.save(ckpt, path)

    @classmethod
    def load(
        cls,
        path,
        map_location=None,
        rf_weight_id: int | None = None,
    ):
        ckpt = torch.load(path, map_location=map_location, weights_only=False)

        if ckpt["kernel_type"] not in [
            "gaussian",
            "poly",
            "linear",
            "biased-gaussian",
            "multiscale-gaussian",
        ]:
            raise ValueError(ckpt["kernel_type"])
        # gnn_kwargs can be in legacy format
        if "gnn" not in ckpt:
            logger.warning(
                "Loading legacy checkpoint with limited "
                "support for extended backbone arguments."
            )
            gnn_kwargs = dict(
                gnn_backbone_id=ckpt["gnn_backbone_id"],
                interaction_block=ckpt["interaction_block"],
            )
        else:
            gnn_kwargs = ckpt["gnn"]["config"]
        model = cls(
            kernel_type=ckpt["kernel_type"],
            random_features_params=ckpt["rf"]["config"],
            jac_chunk_size=ckpt["jac_chunk_size"],
            num_species=ckpt["num_species"],
            **ckpt["input_scaler"]["config"],
            **gnn_kwargs,
        )
        model.rf.load_state_dict(ckpt["rf"]["state_dict"])
        model.input_scaler.load_state_dict(ckpt["input_scaler"]["state_dict"])
        model.energy_shift.load_state_dict(ckpt["energy_shift"])

        if ckpt["multi_weights"] is not None:
            if rf_weight_id is None:
                raise ValueError(
                    f"The checkpoint contains {ckpt['multi_weights'].shape[0]}, select which one to load by specifying rf_weight_id"
                )
            assert rf_weight_id < ckpt["multi_weights"].shape[0]
            model.rf.weights.copy_(
                ckpt["multi_weights"][rf_weight_id].reshape_as(model.rf.weights)
            )

        if map_location is not None:
            return model.to(map_location)
        else:
            return model

    def feature_map(self, data: Configuration):
        """Obtain an embedding of each atom, and map it through
        random features. The RF mapping computes an average so
        the final feature map is per-configuration, instead of
        per-atom.
        """
        gnn_descriptors = self.gnn.descriptors(data)

        normalized_descriptors = self.input_scaler(
            gnn_descriptors,
            atomic_numbers=data.atomic_numbers,
        )
        return self.rf.feature_map(
            normalized_descriptors,
            atomic_numbers=data.atomic_numbers,
        )

    def _feature_map_aux(self, atom_pos: torch.Tensor, data: Configuration):
        old_atom_pos = data.atom_pos
        data.atom_pos = atom_pos
        random_features = self.feature_map(data)
        data.atom_pos = old_atom_pos
        return random_features, random_features

    def grad_feature_map(
        self, data: Configuration
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Compute the feature map for this configuration and
        its gradient with respect to atomic positions

        Returns:
         - forces_fmap : Tensor of size [n_random_features, n_atoms * 3]
         - energy_fmap : Tensor of size [n_random_features]
        """
        if self._grad_fmap_jacfn is None:
            jac_chunk_size = self.get_jacobian_chunk_size(
                self._feature_map_aux, [data.atom_pos, data], argnums=0
            )
            self._grad_fmap_jacfn = jacfwd(
                self._feature_map_aux,
                argnums=0,
                has_aux=True,
                chunk_size=jac_chunk_size,
            )
        return self._grad_fmap_jacfn(data.atom_pos, data)

    def energy(
        self, weights: Optional[torch.Tensor], data: Configuration
    ) -> torch.Tensor:
        r"""Computes the energy of a configuration, using a linear model.

        if :attr:`weights` is not provided, the weights stored in the :attr:`self.rf`
        random features object will be used instead.

        This function returns the energy of the configuration scaled by the number
        of atoms in the configuration itself

        .. math::
            \text{out} = \text{num atoms} \times E(\text{configuration})

        Args:
            weights: The linear coefficients of the energy model.
            configuration: The molecular configuration whose energy to compute.
        """
        if weights is None:
            weights = self.rf.weights

        feature_map = self.feature_map(data).to(dtype=weights.dtype)

        # Contract the last dim of weights with the first of feature_map
        energy = data.natoms * torch.tensordot(weights, feature_map, dims=1)

        return energy

    def _energy_aux(
        self,
        weights: Optional[torch.Tensor],
        atom_pos: torch.Tensor,
        data: Configuration,
    ):
        old_atom_pos = data.atom_pos
        data.atom_pos = atom_pos
        energy = self.energy(weights, data)
        data.atom_pos = old_atom_pos
        return energy, energy

    def grad_energy_func(
        self, weights: Optional[torch.Tensor], data: Configuration
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Computes the gradient of the :meth:`~franken.rf.model.FrankenPotential.energy` acting on a configuration.

        The gradient is equivalent to the negative force acting on
        the configuration.

        if :attr:`weights` is not provided, the weights stored in the :attr:`self.rf`
        random features object will be used instead.

        This function returns a tuple: :attr:`energy_gradient`, :attr:`energy`.

        .. note::
            This function uses the :mod:`torch.func` package to compute the gradient,
            which is particularly useful when computing the energy gradient with
            multiple linear models. In this case :attr:`weights` can be a matrix whose
            first dimension is the number of linear models.
            When computing the gradient for a single linear model, use the
            :meth:`~franken.rf.model.FrankenPotential.grad_energy_autograd` method instead
            for better performance.

        Args:
            weights (Tensor or None): the linear coefficients to compute the energy
            configuration: the molecular configuration whose energy and gradient to compute.

        See also :meth:`~franken.rf.model.FrankenPotential.grad_energy_autograd`.

        """
        if self._grad_energy_jacfn is None:
            jac_chunk_size = self.get_jacobian_chunk_size(
                self._energy_aux, [weights, data.atom_pos, data], argnums=1
            )
            self._grad_energy_jacfn = jacfwd(
                self._energy_aux, argnums=1, has_aux=True, chunk_size=jac_chunk_size
            )
        out = self._grad_energy_jacfn(weights, data.atom_pos, data)
        return out

    def grad_energy_autograd(
        self, weights: Optional[torch.Tensor], data: Configuration
    ) -> Tuple[torch.Tensor | None, torch.Tensor]:
        """Computes the gradient of the :meth:`~franken.rf.model.FrankenPotential.energy` acting on a configuration.

        The gradient is equivalent to the negative force acting on
        the configuration.

        if :attr:`weights` is not provided, the weights stored in the :attr:`self.rf`
        random features object will be used instead.

        This function returns a tuple: :attr:`energy_gradient`, :attr:`energy`. See
        :meth:`~franken.rf.model.FrankenPotential.grad_energy_func` for a discussion on the
        performance characteristics of the two implementations.

        Args:
            weights (Tensor or None): the linear coefficients to compute the energy
            configuration: the molecular configuration whose energy and gradient to compute.
        """
        # Ensure atom positions require gradients
        data.atom_pos.requires_grad_(True)
        # Compute the energy
        energy = self.energy(weights, data)

        if energy.ndim == 0:
            # Scalar case, single gradient
            gradient = torch.autograd.grad(outputs=[energy], inputs=[data.atom_pos])[0]
        else:
            # Vector case, compute gradients for each element independently
            gradients: list[torch.Tensor] = []
            for i in range(energy.shape[0]):
                grad_i = torch.autograd.grad(
                    outputs=[energy[i]],
                    inputs=[data.atom_pos],
                    retain_graph=True,
                )[0]
                assert grad_i is not None
                gradients.append(grad_i)
            # Stack gradients along a new dimension
            gradient = torch.stack(gradients, dim=0)

        return gradient, energy

    def get_jacobian_chunk_size(self, func, func_inputs, argnums=0) -> int:
        if hasattr(self, "_auto_jac_chunk_size"):
            return self._auto_jac_chunk_size
        else:
            jac_chunk_size = self.jac_chunk_size
            if jac_chunk_size == "auto":
                # TODO: We can probably cache the value for different functions to avoid multiple tuner runs.
                jac_chunk_size = tune_jacfwd_chunksize(
                    test_sample=func_inputs,
                    func=func,
                    argnums=argnums,
                    has_aux=True,
                )
                self._auto_jac_chunk_size = jac_chunk_size
                logger.info(
                    f"jacobian chunk size automatically set to {self._auto_jac_chunk_size}"
                )
            assert isinstance(jac_chunk_size, int)
            return jac_chunk_size

    def energy_and_forces(
        self,
        data: Configuration,
        weights: Optional[torch.Tensor] = None,
        forces_mode: str = "torch.autograd",
        add_energy_shift: bool = True,
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        """Infer energy and forces of an atomic configuration using a learned random-features model.

        The parameter :attr:`weights` can be used to specified the model's weights. Otherwise the weights stored in
        :attr:`self.rf.weights` will be used instead.

        The different values of :attr:`forces_mode` correspond to different ways of differentiating
        through the model to obtain the forces acting on the atoms:

        * :code:`"torch.func"` is best for when :attr:`weights` contains multiple linear models on which to perform inference at the same time (in that case :attr:`weights` should be a matrix of shape `[n_linear_models, model_size]`).

        * :code:`"torch.autograd"` is best for when a single linear model is used (i.e. when :attr:`weights` is a vector of shape `[model_size]`)

        * :code:`"no_forces"` can be used if forces are not required.

        Args:
            weights: weights of the random feature model. Defaults to None, in which case the weights set in :attr:`self.rf` will be used instead.

            forces_mode: how to compute the model's forces. Defaults to :code:`"torch.autograd"`.

            add_energy_shift: whether to add the energy shift to the energy.

        Returns:
            energy (Tensor): the scalar (or vector-valued when doing inference with multiple models) valued potential energy.

            forces (Tensor): the forces acting on each atom of the given configuration. If multiple models are given, this will have shape :code:`[n_linear_models, n_atoms, 3]`.
        """
        if forces_mode == "torch.func":
            grad_energy, energy = self.grad_energy_func(weights, data)
            forces = -grad_energy
        elif forces_mode == "torch.autograd":
            grad_energy, energy = self.grad_energy_autograd(weights, data)
            forces = -grad_energy
        elif forces_mode == "no_forces":
            energy = self.energy(weights, data)
            forces = None
        else:
            raise ValueError(f"forces_mode '{forces_mode}' is not valid.")

        if add_energy_shift:
            energy = energy + self.energy_shift(data.atomic_numbers)
        return energy, forces

    def energy_and_forces_from_fmaps(
        self,
        data: Configuration,
        energy_fmap: torch.Tensor,
        forces_fmap: torch.Tensor,
        weights: Optional[torch.Tensor] = None,
        add_energy_shift: bool = True,
    ):
        if weights is None:
            weights = self.rf.weights
        energy = torch.tensordot(
            weights,
            data.natoms * energy_fmap,
            dims=([1], [0]),  # type: ignore
        )
        forces = torch.tensordot(
            weights,
            data.natoms * forces_fmap.view(forces_fmap.shape[0], -1, 3),
            dims=([1], [0]),  # type: ignore
        )
        if add_energy_shift:
            energy = energy + self.energy_shift(data.atomic_numbers)
        return energy, forces

    def forward(
        self,
        data: Configuration,
        weights: Optional[torch.Tensor] = None,
        add_energy_shift: bool = True,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        See docstring of :meth:`~franken.rf.model.FrankenPotential.energy_and_forces`.

        This function defaults to using the 'torch.autograd' strategy which allows the model
        to be jit-compiled.
        """
        grad_energy, energy = self.grad_energy_autograd(weights, data)
        assert grad_energy is not None
        forces = -grad_energy
        if add_energy_shift:
            energy = energy + self.energy_shift(data.atomic_numbers)
        return energy, forces



from typing import Tuple
import typing
import math

import numpy as np
from scipy.stats import linregress


def calc_water_msd(traj):
    """Compute mean squared displacement on the given trajectory

    The MSD is computed at different lag-times, from zero to half the trajectory length,
    such that each MSD is computed averaging over the same number of origins.

    Notes:
    see https://www.utkstair.org/clausius/docs/che548/pdf/selfd.pdf for a discussion
    on practical details with computing the MSD and self-diffusion coefficient

    The algorithm has quadratic complexity in the trajectory length. Faster algorithms
    are possible (see [MDAnalysis](https://docs.mdanalysis.org/stable/documentation_pages/analysis/msd.html)
    for an implementation), but unnecessary unless dealing with very long trajectories.
    """
    t_max = len(traj) // 2
    max_origin = t_max
    msd = np.zeros((t_max, 3))

    # Assume water so MSD is between the oxygen atoms, which are the centers-of-mass of
    # each water molecule in the image
    o_atoms = [atoms[atoms.get_atomic_numbers() == 8] for atoms in traj]
    for origin in range(0, max_origin):
        reference = o_atoms[origin].positions
        for delta_t in range(1, t_max):
            msd[delta_t] += np.mean(
                (o_atoms[origin + delta_t].positions - reference) ** 2, axis=0
            )
    msd /= max_origin
    return msd


def calc_self_diffusion(
    msd, delta_t_fs: float, t_min: int | None = 0, t_min_fs: float | None = None
) -> Tuple[list[float], list[typing.NamedTuple]]:
    """Compute the self-diffusion coefficient from MSD data.

    Args:
        msd (np.ndarray): mean squared displacement as returned by for e.g. :func:`calc_water_msd`
        delta_t_fs (float): the time interval in femtoseconds between successive entries in the MSD
        t_min (int | None): the amount of lag-times to discard from the beginning of msd, mutually
            exclusive with specifying `t_min_fs`.
        t_min_fs (float | None): the amount of lag-times to discard from the beginning of msd
            in femtoseconds. Mutually exclusive with specifying `t_min`.

    Returns:
        Ds (list[float]): the self diffusion coefficient for each coordinate
        linmodels (list[float]): the slopes of the linear fit on the MSD
        intercepts (list[float]): the intercept of the linear fit on the MSD
    """
    t_max = msd.shape[0]

    lagtimes = np.arange(t_max)
    lagtimes_fs = lagtimes * delta_t_fs

    if t_min is None and t_min_fs is not None:
        t_min = int(math.floor(t_min_fs / delta_t_fs))
    elif t_min is None:
        t_min = 0
    linear_models = [
        linregress(lagtimes_fs[t_min:], msd[t_min:, i]) for i in range(msd.shape[1])
    ]
    Ds = [linmod.slope * 1e4 / 2 for linmod in linear_models]
    return Ds, linear_models  # type: ignore
